package MT::Plugin::RevisionControl;
#   RevisionControl - Enable you to control the revisions of entry, webpage and template. You can roll back the content of any time.
#           Programmed by Kunihiro YAMAMOTO (One-X)
#           Copyright (c) 2008 SKYARC System Co.,Ltd.
#           @see http://www.skyarc.co.jp/products/movabletype_plugins/revisioncontrol.html

use strict;
# Use MT's objects
use MT;
use MT::Template;
use MT::Template::Context;
use MT::TemplateMap;
use MT::Plugin;
use MT::Log;
use MT::Entry;
use MT::ObjectTag;
use MT::Serialize;
use MT::Util qw( format_ts offset_time_list encode_html );
# *CAUTION*, not work on MTOS
use CustomFields::Util qw( save_meta get_meta );
# Use Custom MT objects
use MT::RevisionEntry;
use MT::RevisionTemplate;
use MT::RevisionTemplateMap;

use vars qw( $PLUGIN_NAME $VERSION $SCHEMA_VERSION );
$PLUGIN_NAME = 'RevisionControl';
$VERSION = '1.22';
$SCHEMA_VERSION = '1.016';

### Registration of the MT::Plugin
@MT::Plugin::RevisionControl::ISA = qw( MT::Plugin );
my $plugin = __PACKAGE__->new({
        name    => $PLUGIN_NAME,
        version => $VERSION,
        key     => lc $PLUGIN_NAME,
        id      => lc $PLUGIN_NAME,
        author_name => 'SKYARC System Co.,Ltd.',
        author_link => 'http://www.skyarc.co.jp/',
        doc_link => 'http://www.skyarc.co.jp/products/movabletype_plugins/revisioncontrol.html',
        l10n_class => 'RevisionControl::L10N',
        schema_version => $SCHEMA_VERSION,
        description => <<HTMLHEREDOC
<MT_TRANS phrase="Enable you to control the revisions of entry, webpage and template. You can roll back the content of any time.">
HTMLHEREDOC
});
MT->add_plugin( $plugin );

### Registry
sub init_registry {
    my $plugin = shift;
    $plugin->registry({
        object_types => {
            'RevisionEntry' => 'MT::RevisionEntry',             # Entry, WebPage
            'RevisionTemplate' => 'MT::RevisionTemplate',       # Template
            'RevisionTemplateMap' => 'MT::RevisionTemplateMap', # TemplateMap
        },
        callbacks => {
           'MT::App::CMS::template_source.edit_entry' => \&_edit_entry_source,
           'MT::App::CMS::template_param.edit_entry' => \&_edit_entry_param,
           'MT::App::CMS::template_source.preview_strip' => \&_template_param_preview_strip,
           'CMSPostSave.entry' => \&_cms_post_save_entry,
           'CMSPostSave.page' => \&_cms_post_save_entry,
           'MT::Entry::post_remove' => \&_entry_remove,
           'cms_pre_preview' => \&_cms_pre_preview_entry,

           'MT::App::CMS::template_source.edit_template' => \&_edit_template_source,
           'MT::App::CMS::template_param.edit_template' => \&_edit_template_param,
           'MT::App::CMS::template_source.preview_template_strip' => \&_template_param_preview_template_strip,
           'CMSPostSave.template' => \&_cms_post_save_template,
           'MT::Template::post_remove' => \&_template_remove,
           'cms_pre_preview.template' => \&_cms_pre_preview_template,
        },
        methods => {
            rollback_revision_entry => \&_skr_rollback_revision_entry,
            rollback_revision_template => \&_skr_rollback_revision_template,
            clear_all_revisions => \&_skr_clear_all_revisions, # secret command :)
        },
    });
}

### Constant definitions
use constant STATUS_ROLL_BACK => 8;
use constant STATUS_NO_UPDATE => 9;
use constant STATUS_NO_UPDATE_PUBLISH => 10;
use constant REVISION_PUBLISH => 99;
use constant REVISION_FIRST => 1;
use constant MAX_REVISION => 10;



########################################################################
### Template Callback - edit_entry
########################################################################

### Modify the template source in edit_entry
sub _edit_entry_source {
    my( $eh, $app, $tmpl ) = @_;
    my $q = $app->{query};

    #
    my $old = quotemeta( <<'HTMLHEREDOC' );
        <input type="hidden" name="__mode" value="save_entry" />
HTMLHEREDOC
    my $new = <<HTMLHEREDOC;
        <input type="hidden" name="preview_revision" value="" />
HTMLHEREDOC
    $$tmpl =~ s!($old)!$1$new!;

    #
    $old = quotemeta( <<'HTMLHEREDOC' );
    <button
        mt:mode="preview_entry"
        name="preview_entry"
        type="submit"
        accesskey="v"
        title="<mt:var name="preview_button_title">"
        ><__trans phrase="Preview"></button>
HTMLHEREDOC
    $new = <<'HTMLHEREDOC';
    <button
        mt:mode="preview_entry"
        name="preview_entry"
        type="submit"
        onclick="document.forms['entry_form'].preview_revision.value = 0;"
        accesskey="v"
        title="<mt:var name="preview_button_title">"
        ><__trans phrase="Preview"></button>
HTMLHEREDOC
    $$tmpl =~ s!$old!$new!;

    #
    my $entry_id = $q->param( 'id' )
        or return;
    my $entry = MT::Entry->load( $entry_id )
        or return;
    $entry->status != MT::Entry::RELEASE
        and return;

    #
    my $entry_revision = MT::RevisionEntry->load({
            revision => REVISION_FIRST,
            entry_id => $entry_id
    });
    my $active_flg = 0;
    if( $entry_revision ) {
        $active_flg = $entry_revision->active_flg;
        if( $active_flg == 1 || $active_flg == 2 ) {
            my $meta = $entry_revision->data;
            if( $meta ) {
                my $serialize = MT::Serialize->new( 'MT' );
                $meta = $serialize->unserialize( $meta );
                &save_meta( $entry, $$meta );
            }
        } else {
            &_save_revision( $entry, REVISION_PUBLISH, 1 );
        }
    } else {
        &_save_revision( $entry, REVISION_PUBLISH, 1 );
    }

    #
    my $perms = $app->permissions;
    if( $perms->can_publish_post ) {
        $new = '<option value="2"<mt:if name="status_publish"> selected="selected"</mt:if>><__trans phrase="Published"></option>' ;
        $old = quotemeta( $new );

        #
        if( $active_flg == 2 ) {
            $new .= '<option value="'. STATUS_NO_UPDATE_PUBLISH. '" selected="selected" >';
        } else {
            $new .= '<option value="'. STATUS_NO_UPDATE_PUBLISH. '" >';
        }
        $new .= $plugin->translate( 'Not update to release date' ). '</option>' ;

        #
        if( $active_flg == 1 ) {
            $new .= '<option value="'. STATUS_NO_UPDATE. '" selected="selected" >';
        } else {
            $new .= '<option value="'. STATUS_NO_UPDATE. '" >';
        }
        $new .= $plugin->translate( 'Not update publish' ). '</option>' ;

        # 
        if( $active_flg == 1 || $active_flg == 2 ) {
            $new .= '<option value="'. STATUS_ROLL_BACK. '" >'. $plugin->translate( 'Publish roll back' ). '</option>';
        }
        $$tmpl =~ s!$old!$new!;

        #
        if( $active_flg == 1 || $active_flg == 2 ) {
            $old = quotemeta( '<option value="1"<mt:if name="status_draft"> selected="selected"</mt:if>>' );
            $new = '<option value="1">';
            $$tmpl =~ s!$old!$new!g;

            $old = quotemeta( '<option value="2"<mt:if name="status_publish"> selected="selected"</mt:if>>' );
            $new = '<option value="2">';
            $$tmpl =~ s!$old!$new!g;

            $old = quotemeta( '<option value="3"<mt:if name="status_review"> selected="selected"</mt:if>>' );
            $new = '<option value="3">';
            $$tmpl =~ s!$old!$new!g;

            $old = quotemeta( '<option value="4"<mt:if name="status_future"> selected="selected"</mt:if>>' );
            $new = '<option value="4">';
            $$tmpl =~ s!$old!$new!g;
        }
    } else { # !$perms->can_publish_post
        if( $entry->status == 1 ) {
            $old = quotemeta( '<span><__trans phrase="Unpublished (Draft)"></span>' );
        } elsif( $entry->status == 2 ) {
            $old = quotemeta( '<span><__trans phrase="Published"></span>' );
        } elsif( $entry->status == 3 ) {
            $old = quotemeta( '<span><__trans phrase="Scheduled"></span>' );
        } elsif( $entry->status == 4 ) {
            $old = quotemeta( '<span><__trans phrase="Unpublished (Review)"></span>' );
        }

        if( $active_flg == 1 ) {
            $new = '<span>'. $plugin->translate( 'Not update publish' ). '</span>';
            $$tmpl =~ s!$old!$new!g;
        } elsif ($active_flg == 2 ) {
            $new = '<span>'. $plugin->translate( 'Not update to release date' ). '</span>';
            $$tmpl =~ s!$old!$new!g;
        }
    }
}

### Modify the template paramaeter in edit_entry
sub _edit_entry_param {
    my( $eh, $app, $params, $tmpl ) = @_;
    my $q = $app->{query};

    # リビジョンコメント追加
    my $innerHTML = <<'HTMLHEREDOC';
<div class="textarea-wrapper">
    <input name="revision_comment" id="revision_comment" class="full-width" value="" maxlength="255" tabindex="9" mt:watch-change="1" autocomplete="off" />
</div>
HTMLHEREDOC
    my $node = $tmpl->createElement( 'app:setting', {
            id => 'revision_comment',
            label => $plugin->translate( 'Revision comment' ),
            required => 0,
            label_class => 'top-label' });
    $node->innerHTML( $innerHTML );
    my $insert_point = $tmpl->getElementById( 'keywords' );
    $tmpl->insertAfter( $node, $insert_point );

    my $entry_id = $q->param( 'id' )
        or return;
    my $entry = MT::Entry->load( $entry_id )
        or return;
    my $blog =  MT::Blog->load( $entry->blog_id )
        or return;

    $innerHTML = <<'HTMLHEREDOC';
<script type="text/javascript">
    function prev_revision(_revision) {
        document.forms["entry_form"].__mode.value = "preview_entry";
        document.forms["entry_form"].preview_revision.value = _revision;
        document.forms["entry_form"].submit();
        return false;
    }
</script>
<ul>
HTMLHEREDOC

    #
    my $idx = 1;
    my $datas = MT::RevisionEntry->load_iter({ entry_id => $entry->id }, { sort => 'created_on', direction => 'descend' })
        or return;
    while( my $revision = $datas->()) {
        if( $revision->revision ne '99' ){
            my $label_date;
            if( $revision->status == MT::Entry::HOLD()) {
                $label_date = '<span class="icon-left-wide icon-draft">';
            } elsif( $revision->status == MT::Entry::RELEASE()) {
                $label_date = '<span class="icon-left-wide icon-success">';
            } elsif( $revision->status == MT::Entry::FUTURE()) {
                $label_date = '<span class="icon-left-wide icon-future">';
            } else {
                $label_date = '<span class="icon-left-wide">';
            }
            $label_date .= '<a href="javascript:void(0);" onclick="return prev_revision('. $revision->revision. ');" >';
            $label_date .= sprintf '%02d.&nbsp;%s', $idx,
                    format_ts( '%Y-%m-%d %H:%M:%S', $revision->created_on, $blog, $app->user ? $app->user->preferred_language : undef );
            $label_date .= '</a>';
            $label_date .= '</span>';
            if( $revision->revision_comment ){
                $label_date .= '<span class="icon-left-wide" style="display:block; word-break: break-all;" >'. encode_html( $revision->revision_comment ). '</span>';
            }
            $innerHTML .= "<li>$label_date</li>\n";
            $idx++;
        }
    }
    $innerHTML .= "</ul>\n";
    $node = $tmpl->createElement( 'app:setting', {
            id => 'skr_revision',
            label => $plugin->translate( 'Revisions' ),
            required => 0 });
    $node->innerHTML( $innerHTML );
    $insert_point = $tmpl->getElementById( 'entry-publishing-widget' );
    $tmpl->insertAfter( $node, $insert_point );

    #
    my $max_revision = 0;
    my $entry_revision = MT::RevisionEntry->load_iter({ entry_id => $entry->id})
        or return;
    while( my $revision = $entry_revision->()) {
        if( $max_revision < $revision->revision ) {
            if( $revision->revision ne '99' ) {
                $max_revision = $revision->revision;
            }
        }
    }

    #
    if( 0 < $max_revision ) {
        my $rev_entry = MT::RevisionEntry->load({ revision => REVISION_FIRST, entry_id => $entry->id });
        if( $rev_entry ) {
            if( $rev_entry->active_flg == 1 || $rev_entry->active_flg == 2 ) {
                $params->{status} = $rev_entry->status;
                $params->{author_id} = $rev_entry->author_id;
                $params->{allow_comments} = $rev_entry->allow_comments;
                $params->{title} = $rev_entry->title;
                $params->{excerpt} = $rev_entry->excerpt;
                $params->{text} = $rev_entry->text;
                $params->{text_more} = $rev_entry->text_more;
                $params->{convert_breaks} = $rev_entry->convert_breaks;
                $params->{to_ping_urls} = $rev_entry->to_ping_urls;
                $params->{allow_pings} = $rev_entry->allow_pings;
                $params->{keywords} = $rev_entry->keywords;
                $params->{tangent_cache} = $rev_entry->tangent_cache;
                $params->{basename} = $rev_entry->basename;
                $params->{atom_id} = $rev_entry->atom_id;
                $params->{template_id} = $rev_entry->template_id;
                $params->{meta} = $rev_entry->meta;
                $params->{pinged_urls} = $rev_entry->pinged_urls;
                my $authored_on = $rev_entry->authored_on;
                $params->{authored_on} = $authored_on;
                $params->{authored_on_date} = sprintf '%04d-%02d-%02d', ( $authored_on =~ /(\d{4})(\d{2})(\d{2})/ );
                $params->{authored_on_time} = sprintf '%02d:%02d:%02d', ( $authored_on =~ /\d{8}(\d{2})(\d{2})(\d{2})/ );
                $params->{tags} = $rev_entry->tag_str;

                # 
                my $rev_pub = MT::RevisionEntry->load({
                        revision => REVISION_PUBLISH,
                        entry_id => $entry_id });
                if( $rev_pub ) {
                    my $serialize = MT::Serialize->new( 'MT' );
                    my $meta = $serialize->unserialize( $rev_pub->data );
                    &save_meta( $entry, $$meta );
                }
            }
        }
    }
}

### Modify the template parameters in previewing of entry
sub _template_param_preview_strip {
    my( $cb, $app, $tmpl ) = @_;
    my $q = $app->{query};
    my $revision = $q->param( 'preview_revision' )
        or return;
    my $entry_id = $q->param( 'id' );
    my $blog_id = $q->param( 'blog_id' );
    my $type = $q->param( '_type' );

    #
    my $old = quotemeta( <<'HTML' );
                        <button
                            mt:mode="save_entry"
                            type="submit"
                            name="save"
                            value="save"
                            accesskey="s"
                            title="<$mt:var name="save_button_title"$>"
                            class="primary-button"
                            ><$mt:var name="save_button_value"$></button>
                        <button
                            mt:mode="view"
                            type="submit"
                            name="reedit"
                            value="reedit"
                            accesskey="e"
                            title="<$mt:var name="edit_button_title"$>"
                            class="primary-button"
                            ><$mt:var name="edit_button_value"$></button>
HTML
    my $title_yes = $plugin->translate( 'Yes' );
    my $title_no = $plugin->translate( 'No' );
    my $url = "?__mode=rollback_revision_entry&revision=$revision&_type=$type&id=$entry_id&blog_id=$blog_id";
    my $new = <<"HTML";
                        <button
                            mt:mode="rollback_revision_entry"
                            type="button"
                            name="rollback_revision_entry"
                            value="rollback_revision_entry"
                            accesskey="s"
                            title="$title_yes"
                            class="primary-button"
                            onclick="location.href = '$url';"
                            >$title_yes</button>
                        <button
                            mt:mode="rollback_revision_entry"
                            type="button"
                            name="rollback_revision_entry"
                            value="rollback_revision_entry"
                            accesskey="s"
                            title="$title_no"
                            class="primary-button"
                            onclick="history.back(); return false;"
                            >$title_no</button>
HTML
    $$tmpl =~ s!$old!$new!;

    #
    my $msg = $plugin->translate( 'You are about to roll back the revision. Are you sure to execute it ?' );
    $old = quotemeta( <<'HTML' );
            <div id="preview-label"><mt:if name="object_type" eq="entry"><__trans phrase="You are previewing the entry titled &ldquo;[_1]&rdquo;" params="<$mt:var name="title" remove_html="1" escape="html"$>"><mt:else><__trans phrase="You are previewing the page titled &ldquo;[_1]&rdquo;" params="<$mt:var name="title" remove_html="1" escape="html"$>"></mt:if></div>
HTML
    $new = <<"HTML";
            <div id="preview-label">$msg</div>
HTML
    $$tmpl =~ s!$old!$new!;
}

########################################################################
### Function Callbacks - Entry, WePage
########################################################################

### Invoke after saving the entry and webpage
sub _cms_post_save_entry {
    my( $eh, $app, $entry, $original ) = @_;
    my $q = $app->{'query'};

    my $can_publish_post = 0;
    my $perms = $app->permissions;
    if( $perms ) {
        $can_publish_post = $perms->can_publish_post;
    }
    my $entry_revision = MT::RevisionEntry->load_iter({ entry_id => $entry->id })
        or return;
    while( my $rev_entry = $entry_revision->()) {
        my $revision = $rev_entry->revision || 0;
        if( MAX_REVISION < $revision + 1 ) {
            if( $revision ne '99' ) {
                $rev_entry->remove;
            }
        } else {
            $revision = $revision + 1;
            $rev_entry->revision( $revision );
            $rev_entry->save;
        }
    }
    my $revision_comment = $q->param( 'revision_comment' );
    &_save_revision( $entry, REVISION_FIRST, $can_publish_post, $revision_comment );

    my $rev_entry_first = MT::RevisionEntry->load({ revision => REVISION_FIRST, entry_id => $entry->id });
    if( $rev_entry_first ) {
        unless( $rev_entry_first->active_flg == 0 ) {
            my $rev_publish = MT::RevisionEntry->load({ revision => REVISION_PUBLISH, entry_id => $entry->id });
            if( $rev_publish ) {
                # for Each columns
                $entry->blog_id( $rev_publish->blog_id );
                $entry->status( $rev_publish->status );
                $entry->author_id( $rev_publish->author_id );
                $entry->allow_comments( $rev_publish->allow_comments );
                $entry->title( $rev_publish->title );
                $entry->excerpt( $rev_publish->excerpt );
                $entry->text( $rev_publish->text );
                $entry->text_more( $rev_publish->text_more );
                $entry->convert_breaks( $rev_publish->convert_breaks );
                $entry->to_ping_urls( $rev_publish->to_ping_urls );
                $entry->pinged_urls( $rev_publish->pinged_urls );
                $entry->allow_pings( $rev_publish->allow_pings );
                $entry->keywords( $rev_publish->keywords );
                $entry->tangent_cache( $rev_publish->tangent_cache );
                $entry->basename( $rev_publish->basename );
                $entry->atom_id( $rev_publish->atom_id );
                $entry->authored_on( $rev_publish->authored_on );
                $entry->week_number( $rev_publish->week_number );
                $entry->template_id( $rev_publish->template_id );

                # for Tags
                my $tag_delim = chr( $entry->author->entry_prefs->{tag_delim}) || ',';
                my @tags = MT::Tag->split( $tag_delim, $rev_publish->tag_str );
                $entry->tags( @tags );
                $entry->save;

                # for Meta data
                my $serialize = MT::Serialize->new( 'MT' );
                my $meta = $serialize->unserialize( $rev_publish->data );
                &save_meta($entry, $$meta);
            }
            if( $rev_entry_first->active_flg == 3 ){
                $rev_entry_first->active_flg( 0 );
                $rev_entry_first->save;
            }
        }
    }
}

### Invoke before removing the entry and webpage
sub _entry_remove {
    my( $cb, $obj ) = @_;
    my $entry_id = $obj->id;
    my $entry_revision = MT::RevisionEntry->load_iter({ entry_id => $entry_id })
        or return;
    while( my $revision = $entry_revision->()) {
        $revision->remove;
    }
}

### Invoke in previewing of entry
sub _cms_pre_preview_entry {
    my( $cb, $app, $entry, $data ) = @_;
    my $q = $app->{query};
    my $revision = $q->param( 'preview_revision' );
    my $entry_id = $q->param( 'id' );

    if( $revision ) {
        my $entry_revision = MT::RevisionEntry->load({ entry_id => $entry_id, revision => $revision })
            or return;
        # Copy for each columns
        $entry->blog_id( $entry_revision->blog_id );
        $entry->status( $entry_revision->status );
        $entry->author_id( $entry_revision->author_id );
        $entry->allow_comments( $entry_revision->allow_comments );
        $entry->title( $entry_revision->title );
        $entry->excerpt( $entry_revision->excerpt );
        $entry->text( $entry_revision->text );
        $entry->text_more( $entry_revision->text_more );
        $entry->keywords( $entry_revision->keywords );
        $entry->authored_on( $entry_revision->authored_on );
        $entry->week_number( $entry_revision->week_number );
        $entry->convert_breaks( $entry_revision->convert_breaks );
        $entry->to_ping_urls( $entry_revision->to_ping_urls );
        $entry->pinged_urls( $entry_revision->pinged_urls );
        $entry->allow_pings( $entry_revision->allow_pings );
        $entry->atom_id( $entry_revision->atom_id );
        $entry->template_id( $entry_revision->template_id );
        # no updating
        # $entry->tangent_cache( $entry_revision->tangent_cache );
        # $entry->basename( $entry_revision->basename );

        # for Meta data
        my $serialize = MT::Serialize->new( 'MT' );
        my $meta = $serialize->unserialize( $entry_revision->data );
        &save_meta( $entry, $$meta );
    }
}



########################################################################
### Template Callback - edit_template
########################################################################

### Modify the template source in edit_template
sub _edit_template_source {
    my( $eh, $app, $tmpl ) = @_;
    my $q = $app->{query};

    #
    my $old = quotemeta( <<'HTML' );
<input type="hidden" name="__mode" value="save" />
HTML
    my $new = <<'HTML';
<input type="hidden" name="preview_revision" value="" />
HTML
    $$tmpl =~ s!($old)!$1$new!;

    #
    $old = quotemeta( <<'HTML' );
    <button
        onclick="return validate(this.form);"
        type="submit"
        accesskey="s"
        tabindex="12"
        title="<__trans phrase="Save (s)">"
        class="primary-button"
        ><__trans phrase="Save"></button>
    <mt:if name="can_preview">
    <button
        mt:mode="preview_template"
        name="preview_template"
        type="submit"
        accesskey="v"
        tabindex="13"
        title="Preview"
        onclick="return validate(this.form);"
        ><__trans phrase="Preview"></button>
HTML
    $new = <<'HTML';
    <button
        onclick="this.form.__mode.value='save'; return validate(this.form);"
        type="submit"
        accesskey="s"
        tabindex="12"
        title="<__trans phrase="Save (s)">"
        class="primary-button"
        ><__trans phrase="Save"></button>
    <mt:if name="can_preview">
    <button
        mt:mode="preview_template"
        name="preview_template"
        type="submit"
        accesskey="v"
        tabindex="13"
        title="Preview"
        onclick="this.form.preview_revision.value=0; return validate(this.form);"
        ><__trans phrase="Preview"></button>
HTML
    $$tmpl =~ s!$old!$new!;
}

### Modify the template parameters in edit_template
sub _edit_template_param {
    my( $eh, $app, $params, $tmpl ) = @_;
    my $q = $app->{query};
    my $type = $q->param( '_type' );
    my $mode = $q->param( '__mode' );
    my $blog_id = $q->param( 'blog_id' );
    my $revision_id = $q->param('revision');

    # リビジョンコメント追加
    my $node = $tmpl->createElement( 'app:setting', {
            id => 'revision_comment',
            label => $plugin->translate( 'Revision comment' ),
            required => 0,
            label_class => "top-label" });
    $node->innerHTML( <<'HTMLHEREDOC' );
<div class="textarea-wrapper">
   <input name="revision_comment" id="revision_comment" class="full-width" value="" maxlength="255" tabindex="9" mt:watch-change="1"  autocomplete="off"/>
</div>
HTMLHEREDOC
    my $insert_point = $tmpl->getElementById( 'template-body' );
    $tmpl->insertAfter( $node, $insert_point);

    #
    my $template_id = $params->{id}
        or return;
    my $template = MT::Template->load( $template_id )
        or return;
    my $blog = MT::Blog->load( $template->blog_id );
    my $type = $template->type;
    my $mt_uri = $app->mt_uri;

    my $datas = MT::RevisionTemplate->load_iter({ template_id => $template->id },
            { sort => 'created_on', direction => 'descend' })
        or return;
    my $idx = 1;

    my $javascript_action = "";
    if (($type eq 'archive') || ($type eq 'individual') || ($type eq 'category') || ($type eq 'page') || ($type eq 'index')){
        $javascript_action = "prev_revision";
    }else{
        $javascript_action = "load_revision";
        if ($revision_id){
            my $rollback_data = MT::RevisionTemplate->load({ template_id => $template->id, revision => $revision_id  });
            if ($rollback_data){
                $params->{name} = $rollback_data->name;
                $params->{text} = $rollback_data->text;
                $params->{type} = $rollback_data->type;
                $params->{outfile} = $rollback_data->outfile;
                $params->{linked_file} = $rollback_data->linked_file;
                $params->{linked_file_mtime} = $rollback_data->linked_file_mtime;
                $params->{linked_file_size} = $rollback_data->linked_file_size;
                $params->{rebuild_me} = $rollback_data->rebuild_me;
                $params->{build_dynamic} = $rollback_data->build_dynamic;
                $params->{identifier} = $rollback_data->identifier;
                $params->{build_type} = $rollback_data->build_type;
                $params->{build_interval} = $rollback_data->build_interval;
            }
        }
    }
    my $innerHTML = <<HTMLHEREDOC;
<script type="text/javascript">
    function prev_revision( _revision ) {
        document.forms["template-listing-form"].__mode.value = "preview_template";
        document.forms["template-listing-form"].preview_revision.value = _revision;
        document.forms["template-listing-form"].submit();
        return false;
    }
    function load_revision( _revision ) {
        location.href = "$mt_uri?__mode=view&_type=template&id=$template_id&blog_id=$blog_id&revision=" + _revision;
    }
</script>
<ul>
HTMLHEREDOC
    while( my $revision = $datas->()) {
        if( $revision->revision ne '99' ) {
            my $label_date;
            unless( $revision->build_type ) {
                $label_date = '<span class="icon-left icon-neutral" >';
            } else {
                $label_date = '<span class="icon-left icon-publish" >';
            }
            $label_date .= '<a href="javascript:void(0);" onclick="return ' . $javascript_action . '('. $revision->revision. ');" >';
            $label_date .= sprintf '%02d.&nbsp;%s', $idx,
                    format_ts( '%Y-%m-%d %H:%M:%S', $revision->created_on, $blog, $app->user ? $app->user->preferred_language : undef );
            $label_date .= '</a>';
            $label_date .= '</span>';
            if( $revision->revision_comment ) {
                $label_date .= '<span class="icon-left" style="display: block;">'. encode_html( $revision->revision_comment ). '</span>';
            }
            $innerHTML .= "<li>$label_date</li>\n";
            $idx++;
        }
    }
    $innerHTML .= "</ul>\n";
    $node = $tmpl->createElement( 'app:widget', {
            id => 'skr_revision',
            label => $plugin->translate( 'Revisions' )});
    $node->innerHTML( $innerHTML );
    $insert_point = $tmpl->getElementById( 'useful-links' );
    $tmpl->insertBefore( $node, $insert_point );
}

### Modify the template parameters in previewing of entry
sub _template_param_preview_template_strip {
    my( $cb, $app, $tmpl ) = @_;
    my $q = $app->{query};
    my $revision = $q->param( 'preview_revision' )
        or return;
    my $entry_id = $q->param( 'id' );
    my $blog_id = $q->param( 'blog_id' );
    my $type = $q->param( '_type' );

    #
    my $old = quotemeta( <<'HTML' );
                        <button
                            mt:mode="save"
                            type="submit"
                            name="save"
                            value="save"
                            accesskey="s"
                            title="<__trans phrase="Save this template (s)">"
                            class="primary-button"
                            ><__trans phrase="Save this template"></button>
                        <button
                            mt:mode="view"
                            type="submit"
                            name="reedit"
                            value="reedit"
                            accesskey="e"
                            title="<__trans phrase="Re-Edit this template (e)">"
                            class="primary-button"
                            ><__trans phrase="Re-Edit this template"></button>
HTML
    my $title_yes = $plugin->translate( 'Yes' );
    my $title_no = $plugin->translate( 'No' );
    my $url = "?__mode=rollback_revision_template&revision=$revision&_type=$type&id=$entry_id&blog_id=$blog_id";
    my $new = <<"HTML";
                        <button
                            mt:mode="rollback_revision_template"
                            type="button"
                            name="rollback_revision_template"
                            value="rollback_revision_template"
                            accesskey="s"
                            title="$title_yes"
                            class="primary-button"
                            onclick="location.href = '$url';"
                            >$title_yes</button>
                        <button
                            mt:mode="rollback_revision_template"
                            type="button"
                            name="rollback_revision_template_cancel"
                            value="rollback_revision_template_cancel"
                            accesskey="s"
                            title="$title_no"
                            class="primary-button"
                            onclick="history.back(); return false;"
                            >$title_no</button>
HTML
    $$tmpl =~ s!$old!$new!;

    #
    my $msg = $plugin->translate( 'You are about to roll back the revision. Are you sure to execute it ?' );
    $old = quotemeta( <<'HTML' );
            <div id="preview-label">
                <__trans phrase="You are previewing the template named &ldquo;[_1]&rdquo;" params="<$mt:var name="name" escape="html"$>">
                <mt:if name="build_time"><small><__trans phrase="(Publish time: [_1] seconds)" params="<mt:var name="build_time">"></small></mt:if>
            </div>
HTML
    $new = <<HTML;
            <div id="preview-label">$msg</div>
HTML
    $$tmpl =~ s!$old!$new!;
}

########################################################################
### Function Callback - Template
########################################################################

### Invoke after saving the template
sub _cms_post_save_template {
    my( $eh, $app, $template, $original ) = @_;
    my $q = $app->{query};


    my $template_revisions = MT::RevisionTemplate->load_iter({ template_id => $template->id })
        or return;
    while( my $tmp_revision = $template_revisions->()) {
        my $revision = $tmp_revision->revision || 0;
        if( MAX_REVISION < $revision + 1 ){
            $tmp_revision->remove;
        } else {
            $revision = $revision + 1;
            $tmp_revision->revision( $revision );
            $tmp_revision->save;
        }
    }

    my $template_map = MT::RevisionTemplateMap->load_iter({ template_id => $template->id })
        or return;
    while( my $map = $template_map->()) {
        my $revision = $map->revision || 0;
        if( MAX_REVISION < $revision + 1 ) {
            $map->remove;
        } else {
            $revision = $revision + 1;
            $map->revision( $revision );
            $map->save;
        }
    }

    my $revision_comment = $q->param( 'revision_comment' );
    &_revision_save_template( $template, REVISION_FIRST, $revision_comment );
}

### Save the columns for revisions
sub _revision_save_template {
    my( $template, $revision, $revision_comment ) = @_;
    my $template_revision = MT::RevisionTemplate->load({ revision =>$revision ,template_id => $template->id });
    unless( $template_revision ) {
        $template_revision = MT::RevisionTemplate->new;
    }
    $template_revision->template_id( $template->id );
    $template_revision->blog_id( $template->blog_id );
    $template_revision->name( $template->name );
    $template_revision->type( $template->type );
    $template_revision->outfile( $template->outfile );
    $template_revision->text( $template->text );
    $template_revision->linked_file( $template->linked_file );
    $template_revision->linked_file_mtime( $template->linked_file_mtime );
    $template_revision->linked_file_size( $template->linked_file_size );
    $template_revision->rebuild_me( $template->rebuild_me );
    $template_revision->build_dynamic( $template->build_dynamic );
    $template_revision->identifier( $template->identifier );
    $template_revision->build_type( $template->build_type );
    $template_revision->build_interval( $template->build_interval );
    $template_revision->revision( $revision );
    $template_revision->revision_comment( $revision_comment );
    unless( $template_revision->save()) {
        my $msg = $plugin->translate( '[_1]: error template revision data ID:([_2]) [_3]',
                $PLUGIN_NAME, $template->id, $template->name );
        _save_log( $msg, $template->blog_id, 0, MT::Log::ERROR );
        return 0;
    }

    my $template_map = MT::RevisionTemplateMap->load_iter({ template_id => $template->id });
    while( my $map = $template_map->()) {
        $map->remove if $map->revision == $revision;
    }

    my @tmpl_maps = MT::TemplateMap->load({ template_id => $template->id });
    foreach my $map ( @tmpl_maps ) {
        my $revison_map = MT::RevisionTemplateMap->new;
        $revison_map->blog_id( $map->blog_id );
        $revison_map->template_id( $map->template_id );
        $revison_map->archive_type( $map->archive_type );
        $revison_map->file_template( $map->file_template );
        $revison_map->is_preferred( $map->is_preferred );
        $revison_map->build_type( $map->build_type );
        $revison_map->build_interval( $map->build_interval );
        $revison_map->revision( $revision );
        $revison_map->save;
   }
   return 1;
}

### Invoke before removing the template
sub _template_remove {
    my( $cb, $obj ) = @_;
    my $template_id = $obj->id;
    my $templates = MT::RevisionTemplate->load_iter({ template_id => $template_id })
        or return;
    while( my $tmpl = $templates->()) {
        $tmpl->remove;
    }
    my $templatemap = MT::RevisionTemplateMap->load_iter({ template_id => $template_id })
        or return;
    while( my $map = $templatemap->()) {
        $map->remove;
    }
}

### Modify the template parameters in previewing of template
sub _cms_pre_preview_template {
    my( $cb, $app, $template, $data ) = @_;
    my $q = $app->{query};
    my $revision = $q->param( 'preview_revision' )
        or return;
    my $template_id = $q->param( 'id' );

    my $template_revision = MT::RevisionTemplate->load({ revision => $revision ,template_id => $template_id })
        or return;
    $template->blog_id( $template_revision->blog_id );
    $template->name( $template_revision->name );
    $template->type( $template_revision->type );
    $template->outfile( $template_revision->outfile );
    $template->text( $template_revision->text );
    $template->linked_file( $template_revision->linked_file );
    $template->linked_file_mtime( $template_revision->linked_file_mtime );
    $template->linked_file_size( $template_revision->linked_file_size );
    $template->rebuild_me( $template_revision->rebuild_me );
    $template->build_dynamic( $template_revision->build_dynamic );
    $template->identifier( $template_revision->identifier );
    $template->build_type( $template_revision->build_type );
    $template->build_interval( $template_revision->build_interval );
}



########################################################################
### Application Methods - rollback_revision_entry
########################################################################

### Invoke on __mode=rollback_revision_entry
sub _skr_rollback_revision_entry {
    my $app = shift;
    my $q = $app->{query};

    my $revision = $q->param( 'revision' )
        or return $app->error( $plugin->translate( 'select revision error' ));

    my $entry_id = $q->param('id');
    my $entry = MT::Entry->load( $entry_id )
        or return $app->error( $plugin->translate( 'select entry error' ));

    my $entry_revision = MT::RevisionEntry->load({ entry_id => $entry_id, revision => $revision })
        or return $app->error( $plugin->translate( 'revision data not find' ));

    _publish_revision( $plugin, $entry_revision )
        or return $app->error( $plugin->translate( 'rollback error' ));

    my $datas = MT::RevisionEntry->load_iter({ entry_id => $entry->id},
            { sort => 'created_on', direction => 'ascend' })
        or return $app->error( $plugin->translate( 'revision data not find' ));
    while( my $data = $datas->()) {
        next if $data->revision == 99;
        # $data->remove if $data->revision == $revision;
    }

    my $type = $q->param('_type');
    return $app->redirect( $app->uri(
            mode => 'view',
            args => { blog_id => $entry->blog_id, '_type' => $type, id => $entry_id }));
}

###
sub _publish_revision {
    my( $plugin, $revision ) = @_;

    my $entry = MT::Entry->load($revision->entry_id)
        or return 0;
    # for Each columns
    $entry->status( $revision->status );
    $entry->blog_id( $revision->blog_id );
    $entry->author_id( $revision->author_id );
    $entry->allow_comments ($entry->allow_comments );
    $entry->title( $revision->title );
    $entry->excerpt( $revision->excerpt );
    $entry->text( $revision->text );
    $entry->text_more( $revision->text_more );
    $entry->convert_breaks( $revision->convert_breaks );
    $entry->to_ping_urls( $revision->to_ping_urls );
    $entry->pinged_urls( $revision->pinged_urls );
    $entry->allow_pings( $revision->allow_pings );
    $entry->keywords( $revision->keywords );
    $entry->tangent_cache( $revision->tangent_cache );
    $entry->basename( $revision->basename );
    $entry->atom_id( $revision->atom_id );
    $entry->authored_on( $revision->authored_on );
    $entry->week_number( $revision->week_number );
    $entry->template_id( $revision->template_id );
    # $entry->category_id( $revision->category_id );

    # for Tags
    my $tag_delim = chr( $entry->author->entry_prefs->{tag_delim}) || ',';
    my @tags = MT::Tag->split( $tag_delim, $revision->tag_str);
    $entry->tags(@tags);

    # for Meta data
    my $serialize = MT::Serialize->new('MT');
    my $meta = $revision->data;
    $meta = $serialize->unserialize( $meta );
    &save_meta( $entry, $$meta );

    unless( $entry->save()) {
        my $msg = $plugin->translate( '[_1]: error entry publish ID:([_2]) [_3]',
                $PLUGIN_NAME, $entry->id, $entry->title );
        _save_log( $msg, $revision->blog_id, $revision->author_id, MT::Log::ERROR );
        return 0;
    }
    my $msg = $plugin->translate( '[_1]: entry publish ID:([_2]) [_3]',
            $PLUGIN_NAME, $entry->id, $entry->title );
    _save_log( $msg, $revision->blog_id, $revision->author_id, MT::Log::INFO );
    return 1;
}

########################################################################
### Application Methods - rollback_revision_template
########################################################################

### Invoke on __mode=rollback_revision_template
sub _skr_rollback_revision_template {
    my $app = shift;
    my $q = $app->{query};
#    my %param;

    my $revision = $q->param('revision')
        or return $app->error( $plugin->translate( 'select revision error' ));

    my $template_id = $q->param('id');
    my $type = $q->param( '_type' );

    my $template = MT::Template->load( $template_id )
        or return $app->error( $plugin->translate( 'select entry error' ));

    my $template_revision = MT::RevisionTemplate->load({ revision => $revision ,template_id => $template->id })
        or return $app->error( $plugin->translate( 'revision data not find' ));

    _revision_load_template( $plugin, $template_revision )
        or return $app->error( $plugin->translate( 'rollback error' ));

    my $datas = MT::RevisionTemplate->load_iter({ template_id => $template->id },
            { sort => 'created_on', direction => 'ascend' })
        or return $app->error( $plugin->translate( 'revision data not find' ));
    my $del_flg = 0;
    while( my $data = $datas->()) {
        next if $data->revision == 99;
        # $data->remove if $data->revision == $revision;
    }

    my $template_map = MT::RevisionTemplateMap->load_iter({ template_id => $template->id},
            { sort => 'created_on', direction => 'ascend' })
        or return $app->error( $plugin->translate( 'revision data not find' ));
    while( my $data = $template_map->()) {
        next if $data->revision == 99;
        next if $data->revision != $revision;

        my $tmpmap = MT::TemplateMap->new;
        $tmpmap->blog_id( $data->blog_id );
        $tmpmap->template_id( $data->template_id );
        $tmpmap->archive_type( $data->archive_type );
        $tmpmap->file_template( $data->file_template );
        $tmpmap->is_preferred( $data->is_preferred );
        $tmpmap->build_type( $data->build_type );
        $tmpmap->build_interval( $data->build_interval );
        $tmpmap->save;
        # $data->remove;
    }
    return $app->redirect( $app->uri(
            mode => 'view',
            args => { blog_id => $template->blog_id, '_type' => $type, id => $template->id }));
}

###
sub _revision_load_template {
    my( $plugin, $revision ) = @_;

    my $template = MT::Template->load($revision->template_id)
        or return 0;
    $template->blog_id( $revision->blog_id );
    $template->name( $revision->name );
    $template->type( $revision->type );
    $template->outfile( $revision->outfile );
    $template->text( $revision->text );
    $template->linked_file( $revision->linked_file );
    $template->linked_file_mtime( $revision->linked_file_mtime );
    $template->linked_file_size( $revision->linked_file_size );
    $template->rebuild_me( $revision->rebuild_me );
    $template->build_dynamic( $revision->build_dynamic );
    $template->identifier( $revision->identifier );
    $template->build_type( $revision->build_type );
    $template->build_interval( $revision->build_interval );
    unless( $template->save()){
        my $msg = $plugin->translate( '[_1]: error template publish ID:([_2]) [_3]',
                $PLUGIN_NAME, $template->id, $template->name );
        _save_log( $msg, $revision->blog_id, 0, MT::Log::ERROR );
        return 0;
    }
    my $msg = $plugin->translate( '[_1]: template publish ID:([_2]) [_3]',
            $PLUGIN_NAME, $template->id, $template->name );
    _save_log( $msg, $revision->blog_id, 0, MT::Log::INFO );
    return 1;
}



########################################################################
### Application Methods - _skr_clear_all_revisions
########################################################################
sub _skr_clear_all_revisions {
    my $app = shift;

    return $app->return_to_dashboard( permission => 1 )
        unless _permission_check();

    # MT::RevisionEntry
    my @obj_memo = ();
    my $iter = MT::RevisionEntry->load_iter();
    while (my $obj = $iter->()) {
        push @obj_memo, $obj;
    }
    foreach (@obj_memo) {
        $_->remove;
    }

    # MT::RevisionTemplate;
    @obj_memo = ();
    $iter = MT::RevisionTemplate->load_iter();
    while (my $obj = $iter->()) {
        push @obj_memo, $obj;
    }
    foreach (@obj_memo) {
        $_->remove;
    }

    # MT::RevisionTemplateMap;
    @obj_memo = ();
    $iter = MT::RevisionTemplateMap->load_iter();
    while (my $obj = $iter->()) {
        push @obj_memo, $obj;
    }
    foreach (@obj_memo) {
        $_->remove;
    }

    return $app->return_to_dashboard( redirect => 1 );
}



########################################################################
### Common Used Functions
########################################################################

### permission check
sub _permission_check {
    my $app = MT->instance;
    $app->user && $app->user->is_superuser;
}

###
sub _save_revision {
    my( $entry, $revision, $can_publish_post, $revision_comment ) = @_;
    my $active_flg = 0;
    my $entry_revision = MT::RevisionEntry->load({ revision => $revision, entry_id => $entry->id });
    unless( $entry_revision ){
        $entry_revision = MT::RevisionEntry->new;
    } else {
        unless( $can_publish_post ) {
            $active_flg = $entry_revision->active_flg;
        }
    }
    $entry_revision->entry_id( $entry->id );
    $entry_revision->revision( $revision );
    $entry_revision->blog_id( $entry->blog_id );
    $entry_revision->author_id( $entry->author_id );
    $entry_revision->allow_comments( $entry->allow_comments );
    $entry_revision->title( $entry->title );
    $entry_revision->excerpt( $entry->excerpt );
    $entry_revision->text( $entry->text );
    $entry_revision->text_more( $entry->text_more );
    $entry_revision->convert_breaks( $entry->convert_breaks );
    $entry_revision->to_ping_urls( $entry->to_ping_urls );
    $entry_revision->pinged_urls( $entry->pinged_urls );
    $entry_revision->allow_pings( $entry->allow_pings );
    $entry_revision->keywords( $entry->keywords );
    $entry_revision->tangent_cache( $entry->tangent_cache );
    $entry_revision->basename( $entry->basename );
    $entry_revision->atom_id( $entry->atom_id );
    $entry_revision->authored_on( $entry->authored_on );
    $entry_revision->week_number( $entry->week_number );
    $entry_revision->template_id( $entry->template_id );
    # $entry_revision->category_id( $entry->category_id );

    my @ts = offset_time_list( time, $entry->blog_id );
    my $ts = sprintf '%04d%02d%02d%02d%02d%02d', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
    $entry_revision->publish_date( $ts );
    $entry_revision->revision_comment( $revision_comment );

    # for Tags
    my $tag_delim = chr( $entry->author->entry_prefs->{tag_delim}) || ',';
    my @tags = $entry->tags;
    my $tag_str = join $tag_delim, @tags;
    $entry_revision->tag_str( $tag_str );

    # for Meta data
    my $meta = get_meta( $entry );
    my $serialize = MT::Serialize->new( 'MT' );
    $meta = $serialize->serialize( \$meta );
    $entry_revision->data( $meta );

    unless( $can_publish_post ) {
        $entry_revision->status( MT::Entry::RELEASE );
        $entry_revision->active_flg( $active_flg );
    } elsif( $entry->status == STATUS_NO_UPDATE ) {
        $entry_revision->status( MT::Entry::RELEASE );
        $entry_revision->active_flg( 1 );
    } elsif( $entry->status == STATUS_NO_UPDATE_PUBLISH ) {
        $entry_revision->status( MT::Entry::RELEASE );
        $entry_revision->active_flg( 2 );
    } elsif( $entry->status == STATUS_ROLL_BACK ){
        $entry_revision->status( MT::Entry::RELEASE );
        $entry_revision->active_flg( 3 );
    } else {
        $entry_revision->status( $entry->status );
        $entry_revision->active_flg( 0 );
    }
    unless( $entry_revision->save()) {
        my $msg = $plugin->translate( '[_1]: save error entry revision ID:([_2]) [_3]',
                $PLUGIN_NAME, $entry_revision->id, $entry_revision->title );
        _save_log( $msg, $entry_revision->blog_id, $entry_revision->author_id, MT::Log::ERROR );
    }
}

### Logging
sub _save_log {
    my( $msg, $blog_id, $author_id, $level ) = @_;

    my $log = MT::Log->new;
    $log->message( $msg );
    $log->blog_id( $blog_id );
    $log->author_id( $author_id );
    $log->level( $level );
    my @ts = gmtime;
    my $ts = sprintf '%04d%02d%02d%02d%02d%02d', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
    $log->modified_on( $ts );
    $log->created_on( $ts );
    $log->save;
}
1;